import copy
import inspect
from concurrent.futures._base import Executor
from typing import Dict, Any, Tuple

KwArgs = Dict[str, Any]


def _inside_jupyter() -> bool:
    # an approach from stackoverflow, has a couple of problems
    # doesn't handle 'jupyter console'
    try:
        shell = get_ipython().__class__.__name__
        if shell == 'ZMQInteractiveShell':
            return True  # Jupyter notebook or qtconsole
        elif shell == 'TerminalInteractiveShell':
            return False  # Terminal running IPython
        else:
            return False  # Other type (?)
    except NameError:
        return False


def _divide_kwargs(kwargs: KwArgs, executor_class: Executor) -> Tuple[KwArgs, KwArgs]:
    executor_args = {
        k: kwargs[k] for k in inspect.getfullargspec(executor_class)[0]
        if k in kwargs
    }
    tqdm_args = copy.copy(kwargs)
    for k in inspect.getfullargspec(executor_class)[0]:
        if k in tqdm_args:
            del tqdm_args[k]

    return executor_args, tqdm_args



from concurrent.futures import ProcessPoolExecutor
from typing import Any, Callable, Iterable, Optional, Union

from bounded_pool_executor import BoundedProcessPoolExecutor
from typing_extensions import Literal

from pqdm._base import _parallel_process


def pqdm(
    array: Iterable[Any],
    function: Callable[[Any], Any],
    n_jobs: int,
    argument_type: Optional[Union[Literal['kwargs'], Literal['args']]] = None,
    bounded: bool = False,
    **kwargs
):
    return _parallel_process(
        iterable=array,
        function=function,
        argument_type=argument_type,
        n_jobs=n_jobs,
        executor=BoundedProcessPoolExecutor if bounded else ProcessPoolExecutor,
        **kwargs
    )



#!/usr/bin/env python

"""Tests for `pqdm` package."""
import io

import pytest

from pqdm.processes import pqdm as pqdm_processes
from pqdm.threads import pqdm as pqdm_threads


def multiply_args(a, b):
    return a * b


def multiply_list(x):
    return x[0] * x[1]


RESULT = [1 * 2, 2 * 3, 3 * 4, 4 * 5]

TEST_DATA = [
    (
        multiply_args,
        [{'a': 1, 'b': 2}, {'a': 2, 'b': 3}, {'a': 3, 'b': 4}, {'a': 4, 'b': 5}],
        {
            'n_jobs': 2,
            'argument_type': 'kwargs',
        }
    ),
    (
        multiply_args,
        [[1, 2], [2, 3], [3, 4], [4, 5]],
        {
            'n_jobs': 2,
            'argument_type': 'args',
        }
    ),
    (
        multiply_list,
        [[1, 2], [2, 3], [3, 4], [4, 5]],
        {
            'n_jobs': 2
        }
    ),

]


@pytest.mark.parametrize("function, input_list, kwargs", TEST_DATA)
def test_pqdm_threads_work_with_argument_types(function, input_list, kwargs):
    result = pqdm_threads(input_list, function, **kwargs)
    assert result == RESULT


@pytest.mark.parametrize("function, input_list, kwargs", TEST_DATA)
def test_pqdm_processes_work_with_argument_types(function, input_list, kwargs):
    result = pqdm_processes(input_list, function, **kwargs)
    assert result == RESULT


@pytest.mark.parametrize("function, input_list, kwargs", TEST_DATA)
def test_pqdm_processes_pushes_argument_to_tqdm(function, input_list, kwargs):
    output = io.StringIO("")

    kwargs['desc'] = 'Testing'
    kwargs['file'] = output

    result = pqdm_processes(input_list, function, **kwargs)

    text = output.getvalue()
    assert 'Testing:' in text
    assert result == RESULT


@pytest.mark.parametrize("function, input_list, kwargs", TEST_DATA)
def test_pqdm_threads_pushes_argument_to_tqdm(function, input_list, kwargs):
    output = io.StringIO("")

    kwargs['desc'] = 'Testing'
    kwargs['file'] = output

    result = pqdm_threads(input_list, function, **kwargs)

    text = output.getvalue()
    assert 'Testing:' in text
    assert result == RESULT



import copy
from concurrent.futures import Executor, as_completed
from typing import Any, Callable, Iterable

from tqdm import tqdm as tqdm_cli
from tqdm.notebook import tqdm as tqdm_notebook

from pqdm.constants import ArgumentPassing
from pqdm.utils import _inside_jupyter, _divide_kwargs

TQDM = tqdm_notebook if _inside_jupyter() else tqdm_cli


def _handle_singular_processor(array, function, argument_type, tqdm_opts):
    if argument_type == ArgumentPassing.AS_KWARGS:
        return [function(**a) for a in TQDM(array, **tqdm_opts)]
    elif argument_type == ArgumentPassing.AS_ARGS:
        return [function(*a) for a in TQDM(array, **tqdm_opts)]
    else:
        return [function(a) for a in TQDM(array, **tqdm_opts)]


def _parallel_process(
    iterable: Iterable[Any],
    function: Callable[[Any], Any],
    n_jobs: int,
    executor: Executor,
    argument_type: str = 'direct',
    **kwargs
):
    executor_opts, tqdm_opts = _divide_kwargs(kwargs, executor)
    executor_opts['max_workers'] = n_jobs

    if n_jobs == 1:
        return _handle_singular_processor(iterable, function, argument_type, tqdm_opts)

    with executor(**executor_opts) as pool:
        if argument_type == ArgumentPassing.AS_KWARGS:
            futures = [pool.submit(function, **a) for a in iterable]
        elif argument_type == ArgumentPassing.AS_ARGS:
            futures = [pool.submit(function, *a) for a in iterable]
        else:
            futures = [pool.submit(function, a) for a in iterable]


        pre_opts = copy.copy(tqdm_opts)
        pre_opts['desc'] = 'SUBMITTING | ' + pre_opts.get('desc','')

        for _ in TQDM(
            as_completed(futures),
            **pre_opts
        ):
            pass

        tqdm_opts['desc'] = 'COLLECTING | ' + tqdm_opts.get('desc', '')

    results = []
    for i, future in TQDM(enumerate(futures), **tqdm_opts):
        try:
            results.append(future.result())
        except Exception as e:
            results.append(e)

    return results
